
struct node
	{
	unsigned int node_num;

	signed int left;   		// links for
	signed int right;  		// daughter nodes

	unsigned long int frequency;		//of occurence in source file

	signed int prev_num, next_num;	//links for headnode linked list
	};

class node_tree
	{
  private:
	unsigned long int old_byte_tot;

	int parent_node;		//num of deepest root in tree-or the topmost node
	int last_node;		// (total nodes - 1), or last node in use

	node* nodes[513];		//513--enough to hold the largest tree

	//long unsigned int prev_bit_tot, new_bit_tot;
									//for calc of compression percentage

	void find_small(int&, int&, int);//find the two smallest in headnode list
	void go_in(unsigned int, char*, char**);  //used during finding codes

  public:
	void read_from_file(char*);		//initializes
	void do_Hauffman();             //algorithm: brains of program
	void form_codes(char**);
	void write_to_file(char*, char*);

	//used for expansion
	int go(int);
	unsigned long int read_tree_from_file(char*);		//initializes

	};

void node_tree::read_from_file(char* filename)
	{
  //handles task of initialization

	static unsigned long int freq_table[256];
		//it is static for automatic initialization to 0
		//for (int i=0; i<=255; i++)   //initialize freq_table array
		//	freq_table[i] = 0;

  //STEP 1: Read from file and fill in the frequencies
  // in temporary freq_table array
	ifstream inp_file;
	inp_file.open(filename, ios::in | ios::binary);

	unsigned char byte;
	while (1)
		{
		inp_file.get(byte);
		if (!inp_file) break; //bcos test reveals eof AFTER (-1) has been read
		freq_table[byte]++;
		}

	inp_file.close();

  //STEP 2: Using the temporary freq_table array, initialize
  // the nodes

	int count=0;
	for (int j=0; j<=255; j++)
		if (freq_table[j])
			{
			nodes[count] = new node;

			nodes[count]->node_num = count;

			nodes[count]->prev_num = count - 1;
			nodes[count]->next_num = count + 1;

			nodes[count]->left = -1;
			nodes[count]->right = j;

			nodes[count]->frequency = freq_table[j];

			count++;
			}

	parent_node = -1;
	last_node = count-1;
	nodes[last_node]->next_num = -1;	//indicating currently last head node

	//prev_bit_tot = new_bit_tot = 0;
	}

void node_tree::do_Hauffman()
  //This routine is the "brains" of the program
  //It implements the Huffman's coding algorithm
	{
	int start_head_node = 0;
	int head_nodes = last_node+1;
	int ind1, ind2;

  //start main loop
	while (head_nodes > 1)
		{
		find_small(ind1, ind2, start_head_node);

		last_node++;
		head_nodes--;

	  // generate a parent node with ind1 and ind2 nodes as daughters

		nodes[last_node] = new node;

		nodes[last_node]->node_num = last_node;

		nodes[last_node]->left = ind1;
		nodes[last_node]->right = ind2;

		nodes[last_node]->frequency=
						nodes[ind1]->frequency + nodes[ind2]->frequency;

	  //insert the new node in place of ind1 node
		nodes[last_node]->prev_num = nodes[ind1]->prev_num;
		if (nodes[ind1]->prev_num != -1)
			nodes[nodes[ind1]->prev_num]->next_num = last_node;
		else
			start_head_node = last_node;
		nodes[last_node]->next_num = nodes[ind1]->next_num;
		if (nodes[ind1]->next_num != -1)
			nodes[nodes[ind1]->next_num]->prev_num = last_node;

	  //pull out ind2 node
		if (nodes[ind2]->prev_num != -1)
			nodes[nodes[ind2]->prev_num]->next_num = nodes[ind2]->next_num;
		else
			start_head_node = nodes[ind2]->next_num;
		if (nodes[ind2]->next_num != -1)
			nodes[nodes[ind2]->next_num]->prev_num = nodes[ind2]->prev_num;

		}
	parent_node = start_head_node;
	}

void node_tree::find_small(int& num1, int& num2, int start_head_node)
	{       //find the two smallest in headnode linked list
	int curnode;
	long unsigned int small;

  //find smallest
	curnode = start_head_node;
	small = nodes[curnode]->frequency; num1 = curnode;
	do
		{
		if (nodes[curnode]->frequency < small)
			{
			small = nodes[curnode]->frequency;
			num1 = curnode;
			}
		curnode = nodes[curnode]->next_num;
		}
	while (curnode != -1);

  //find second smallest
	curnode = (num1 == start_head_node) ?
		(nodes[start_head_node]->next_num) : (start_head_node);
	small = nodes[curnode]->frequency; num2 = curnode;
	do
		{
		if (curnode != num1)
			if (nodes[curnode]->frequency < small)
				{
				small = nodes[curnode]->frequency;
				num2 = curnode;
				}
		curnode = nodes[curnode]->next_num;
		}
	while (curnode != -1);

	}

void node_tree::form_codes(char** codes)
	{

	char temp_string[256] = "";

	go_in(parent_node, temp_string, codes);

	}

void node_tree::go_in(unsigned int cur_node, char* temp, char** codes)
	{

	if (nodes[cur_node]->left == -1)	//terminal node reached
		{
		int a = nodes[cur_node]->right, b = strlen(temp);

		codes[a] = new char[b+1];

		strcpy(codes[a], temp);

		//prev_bit_tot += ( (nodes[cur_node]->frequency) * 8 );
		//new_bit_tot += ( (nodes[cur_node]->frequency) * b );

		return;
		}
	else
		{
		temp[strlen(temp)+1] = 0;		//increase length by 1

		temp[strlen(temp)] = '0';       //insert last digit as 0
		go_in(nodes[cur_node]->left, temp, codes);

		temp[strlen(temp)-1] = '1';     //make it "1"
		go_in(nodes[cur_node]->right, temp, codes);

		temp[strlen(temp)-1] = 0;        //decrease length by 1

		return;
		}
	}

void node_tree::write_to_file(char* src_file, char* out_file_name)
	{
	extern unsigned long int file_length(char*);

	ofstream out_file;
	out_file.open(out_file_name, ios::out | ios::binary);

	//write old length of file
	old_byte_tot = file_length(src_file);
	out_file.write( (char*)&old_byte_tot, 4);

	//write root_node, last_node
	out_file.write( (char*)&parent_node, 2);
	out_file.write( (char*)&last_node, 2);

	//write tree
	for(int i=0;i<=last_node; i++)
		{
		out_file.write( (char*)&nodes[i]->left, 2);
		out_file.write( (char*)&nodes[i]->right, 2);
		}

	out_file.close();
	}

unsigned long int node_tree::read_tree_from_file(char* src_file)
	{

	ifstream in_file;

	in_file.open(src_file, ios::in | ios::binary);

  //read old length of file
	in_file.read( (char*)&old_byte_tot, 4);

  //read root node, last node
	in_file.read( (char*)&parent_node, 2);
	in_file.read( (char*)&last_node, 2);

  //read tree
	for(int i=0;i<=last_node; i++)
		{
		nodes[i] = new node;
		in_file.read( (char*)&nodes[i]->left, 2);
		if (!in_file) return 0;	  //file ends before than that expected
		in_file.read( (char*)&nodes[i]->right, 2);
		}

	unsigned long int last_position = in_file.tellg();

	in_file.close();
	return last_position;  	//successfull!

	}

int node_tree::go(int direction)
	{
	static cur_node = parent_node;
	static unsigned long int num_of_bytes = 0;

	cur_node = (direction == 0) ?
					nodes[cur_node]->left : nodes[cur_node]->right;

	if (nodes[cur_node]->left == -1)
		{	//terminal node
		int answer = nodes[cur_node]->right;	//character represented

		cur_node = parent_node;  //reset current node to the root node

	  //check whhether all bytes has been read
		if ((++num_of_bytes) == old_byte_tot)
			answer += 256;    //indicates last byte has been read
					//actually, it avoids reading the last filler zeroes

        return answer;
		}
	else
		return 1000;		//indicates: terminal node not found

	}
